# This module decodes NWS CCF (Coded Cities Forecast) data

# Copyright 2004 Matthew Miller
#
# This library is free software; you can redistribute it and/or modify
# it under the same terms as Perl itself.

# For example, from
# <http://weather.noaa.gov/pub/data/raw/fp/fpus41.kbox.ccf.bos.txt>
# on March 14, 2004:

#
# FPUS41 KBOX 141011
# CCFBOS
# 
# BOS AB 042/037 050/029 040 09081
#     BESBB 032/042 032/038 034/042 033/049  1231454221
# PVD BB 043/037 051/029 040 09081
#     EEOBB 032/041 032/039 032/045 034/050  1331554211
# BDL BB 044/035 048/026 038 09161
#     EESBB 029/040 030/039 030/044 030/049  1232554212
# EEN BB 042/030 042/022 037 09172
#     BESBB 022/037 025/035 026/037 029/042  1111554222
# MHT BB 042/030 046/025 040 09170
#     BESBB 024/040 027/037 028/039 030/048  1111454222
# ACK AB 043/039 049/033 042 09072
#     EESBB 034/040 033/038 035/042 039/049  1341454211
# CEF BB 043/034 046/025 038 09171
#     BESBB 028/039 029/038 029/042 030/049  1232554212
# HYA AB 042/039 049/030 041 09072
#     EEOBB 032/040 033/039 034/041 033/049  1341454211
# LWM BB 042/036 047/026 040 09071
#     BESBB 027/040 029/039 030/040 033/049  1111454222
# ORE BB 042/031 046/023 039 09172
#     BESBB 026/038 028/038 027/039 029/043  1221554212
# ORH BB 042/034 045/024 037 09081
#     BESBB 026/038 029/037 028/041 031/044  1231554212
# TAN BB 043/037 050/028 041 09081
#     EEOBB 031/041 033/039 032/045 033/051  1241554211
# IJD BB 044/035 047/026 039 09161
#     EESBB 030/041 032/039 031/045 032/050  1332554212
# WST BB 044/038 048/030 040 09151
#     EESBB 033/042 032/039 033/044 034/048  1342554211
#
#

# I have no idea where this format is officially specified, if it is at all,
# but it's somewhat explained at:
# <http://www.crh.noaa.gov/lmk/product_guide/products/forecast/ccf.htm>

# There are other descriptions at various places on various weather.gov and
# noaa.gov sites, but they're often much less helpful and sometimes just
# plain wrong. The two-line format seems to be an extended version of some
# original specification -- kind of annoying, but no match for The Power of
# Perl.

# In short, the format is:

# Sss CC Ttt/Ttt Ttt/Ttt Ttt IiPPP
#     CCCCC Ttt/Ttt Ttt/Ttt Ttt/Ttt Ttt/Ttt Ttt PPPPPPPPPPP

# where Sss is a station code, C is a predicted condition, Ttt is a high or
# low temperature, Ii is the forecaster's id, and P is a probability of
# precipitation. You'll notice that in the example data above, some of the
# fields -- the last Ttt and P -- are missing. Yup. That happens.

# C is for a 24-hour period, and Ttt and P are for 12 hour periods. Starting
# immediately after the forecast is published, I assume, but if you know
# better, let me know.

# The forecasts are produced twice-daily, between 1:30 and 5:30 am and pm,
# local time. To add to the fun, the morning forecast lists temperatures in
# in the order high-low-high-... and the evening one low-high-low-..., and
# there's no good way to tell which is which. Each CCF file is timestamped,
# but it's in GMT. Given that the forecasts are mostly for the US, this
# could be worked around (even though the difference between 5:30 am in
# Hawaii and the east coast on DST at 1:30 is getting pretty small), but
# that feels kinda icky. So instead, I autodetect the order, which seems to
# work pretty well. This also has the advantage of working on partial file
# entries -- you just need the data lines.

# Anyway, without further ado:


package Geo::CCF;


use 5.008;
use strict;
use warnings;


our $VERSION = '0.04';


## Weather Code Lookup Table

my %_weather_codes = ( 
	A => "Fair",
	B => "Partly Cloudy",
	C => "Cloudy",
	D => "Dust",
	E => "Mostly Cloudy",
	F => "Foggy",
	G => "Very Hot",
	H => "Haze",
	I => "Very Cold",
	J => "Snow Showers",
	K => "Smoke",
	L => "Drizzle",
	M => "Snow Showers",
	N => "Windy",
	O => "Rain/Snow Mix",
	P => "Blizzard",
	Q => "Blowing Snow",
	R => "Rain",
	S => "Snow",
	T => "Thunderstorms",
	U => "Sunny",
	V => "Clear",
	W => "Rain Showers",
	X => "Sleet",
	Y => "Freezing Rain",
	Z => "Freezing Drizzle"
);

## Constants. Or rather, just Constant.

# The number of 12-hour periods covered by the normal CCF format at this
# point is pretty much always 14. You don't want to change this unless the
# format really does change, in which case you'll also have to change the
# parsing code below.
use constant NUM_FORECAST_PERIODS => 14;


## Constructor

sub new {
	my $this = shift;
	my $class = ref($this) || $this;
	my $self = {};

	# I wanted to follow the Geo::METAR format as much as possible, but
	# the multiple-cities thing kind of throws that off. Oh well.

	bless $self, $class;
	return $self;
}


## the main function

# take CCF data and process it
sub ccf {
	my $self = shift;

	my $CCF = shift or return $self;

	# get rid of stupid indented two-line format
	$CCF =~ s/\n    / /mg;

	# the slashes in the format are meaningless and bad. destroy them.
	$CCF =~ s/\// /mg;

	# sometimes, fields are just left blank. like, null. yes, that's dumb.
	$CCF =~ s/ (?= )/ ___/g;
	
	# these are whole-file values we may or may not find below.
	my ($ccfcode, $ccfstation, $ccfdate, $ccftime);
	
	# okay, now spit back into lines, and process	
	foreach (split /\n/, $CCF) {

		# detect the header/date line. I don't actually undertand
		# the meaning of the first field. The second is the station
		# from which them report came, and the third is day-of-month
		# hour minute, GMT. We could try to process the timestamp
		# into a real time value, but since we have no idea what
		# month this is, that seems dangerous.
		#
		# we keep looking for new lines like this, in case we have
		# several files concatenated.
		if (/^(FP[A-Z]{2}\d\d) ([A-Z]{4}) (\d{2})(\d{4})$/) 
		{
			($ccfcode, $ccfstation, $ccfdate, $ccftime) =
			 /^(FP[A-Z]{2}\d\d) ([A-Z]{4}) (\d{2})(\d{4})$/
			 and next;
		}

		
		# this crazy thing breaks out the actual lines and puts it 
		# into the-hopefully-more-useful data structure that this
		# module is all about.
		my ($station, @wcode, @temp, $forecasterid, @prop);

		($station, $wcode[0], $wcode[1], 
		 $temp[0], $temp[1], $temp[2], $temp[3], $temp[4],
		 $forecasterid, $prop[0], $prop[1], $prop[2], 
		 $wcode[2], $wcode[3], $wcode[4], $wcode[5], $wcode[6], 
		 $temp[5], $temp[6], $temp[7], $temp[8], $temp[9],
		 $temp[10], $temp[11], $temp[12], $temp[13],
		 $prop[3], $prop[4], $prop[5], $prop[6], $prop[7], $prop[8], 
		 $prop[9], $prop[10], $prop[11], $prop[12], $prop[13]
		) =
		 /^([A-Z]{3})[ ]([A-Z])([A-Z])[ ]
           ([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]
           (\d\d)([\d+-])([\d+-])([\d+-])[ ]
           ([A-Z])([A-Z])([A-Z])([A-Z])([A-Z])[ ]
           ([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]
           ([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]([\d_]{3})[ ]
           ([\d+-])([\d+-])([\d+-])([\d+-])([\d+-])([\d+-])
           ([\d+-])([\d+-])([\d+-])([\d+-])([\d+-]?)/x
		 or next;
		 

		# it may be a bit wasteful to duplicate this, but it's
		# convenient to have it with each one later (perhaps for)
		# mixing data from different files).
		$self->{$station}{CCFCODE}    = $ccfcode    if $ccfcode;
		$self->{$station}{CCFSTATION} = $ccfstation if $ccfstation;
		$self->{$station}{CCFDATE}    = $ccfdate    if $ccfdate;
		$self->{$station}{CCFTIME}    = $ccftime    if $ccftime;


		# and in case this is needed, here it is.
		$self->{$station}{FORECASTERID} = $forecasterid;

		
		# loop through the forecast periods, and store predicted
		# temperature and precipitation-chances.
		for (my $i=0; $i<NUM_FORECAST_PERIODS; $i++) {

			# resolve weather element letters to human-understandable text.
			# int($i/2) makes this double-up, because there's 7 24-hour
			# condition forecasts, but we want to match the 14 12-hour 
			# temperature and precipitation-chance forecasts.		
			$self->{$station}{CONDITION}[$i] = 
			                              $_weather_codes{$wcode[int($i/2)]};

			# decode and store the probability of precipitation
			$self->{$station}{PROBPRECIP}[$i] = _decodeprobprecip($prop[$i]);

		}


		# which hi/lo order does this particular entry use?
		# urgh. see above (and below) for rants on this.
		my $first;
		my $second;
		if (_detectlohiorder(\@temp) eq "lofirst") {
			$first  = 'LOW';
			$second = 'HIGH';
		} else {
			$first  = 'HIGH';
			$second = 'LOW';
		}
	
		# okay, now sort that all out and save hi/lo to useful arrays 
		# as mentioned, there's 14 12-hour periods in the report.
		for (my $i=0; $i<NUM_FORECAST_PERIODS; $i += 2) {

			$self->{$station}{$first}[$i]  = _decodetemp($temp[$i]);
			$self->{$station}{$second}[$i] = undef;
				
			$self->{$station}{$first}[$i+1]  = undef;
			$self->{$station}{$second}[$i+1] = _decodetemp($temp[$i+1]);

		}

	}
		              
	return $self
}

## Here we've got several internal functions for data-crunching


# _detecthiloorder: figure out if this file is for day or night, basically.
#
# The stupid format switches around from day to night. And the file has a
# timestamp at the top, but day/night is a local thing, and the timestamp is
# GMT, so that seems dodgy too. Instead I autodetect -- hi temps should
# generally be higher than lo temps. In the event that this is inverted for
# a whole week, it's probably the apocalypse.

sub _detectlohiorder {
	my $tempref = shift;
	my $loscore = 0;
	my $hiscore = 0;
	
	for (my $i=0; $i<NUM_FORECAST_PERIODS; $i+=2) {
		(${$tempref}[$i]   =~ /^\d+$/) &&
        	(${$tempref}[$i+1] =~ /^\d+$/)
		 or next;

		$hiscore++ if ${$tempref}[$i] > ${$tempref}[$i+1];
		$loscore++ if ${$tempref}[$i] < ${$tempref}[$i+1];
	}
	
	# I guess go with loscore if there's a tie. That'd be weird, though.
	if ($hiscore>$loscore) {
		return "hifirst";
	} else {
		return "lofirst";
	}	
}


# _decodetemp: put temperature values in simple integer format.
#
# the encoding is silly -- negative numbers are +900!

sub _decodetemp {
	my $temp = shift;

	$temp =~ /^\d+$/ or return undef;
                    
	$temp += 0;  # convert string to number;
	$temp = -($temp-900) if $temp >= 900; # handle stupid 900 thing
	return $temp;
}

# _decodeprobprecip: decode probability of precipitation
#
# It's represented like this:
# 0=0%, 1 = 10%, 2 = 20%, ... 9 = 90%; + = 100%, - = 5%
sub _decodeprobprecip {
	my $prop = shift;
	
	$prop =~ /^\d$/ and return $prop*10;
	$prop eq "+"    and return 100;
	$prop eq "-"    and return 5;
	                return undef;

}

## that's all.


1;

__END__

=head1 NAME

Geo::CCF - decode National Weather Service Coded Cities Forecast data

=head1 SYNOPSIS

  use Geo::CCF;
  
  my $forecast = new Geo::CCF;

  my $station = "BOS";

  $forecast->ccf($ccf_data_downloaded_from_noaa);

  die("Can't find info for $station") if !$forecast->{$station};

  for (my $i=0; $i<Geo::CCF::NUM_FORECAST_PERIODS; $i++) {
    print $forecast->{$station}{CONDITION}[$i];
    $forecast->{$station}{HIGH}[$i] and
        print ", high $forecast->{$station}{HIGH}[$i]";
    $forecast->{$station}{LOW}[$i] and
        print ", low $forecast->{$station}{LOW}[$i]";
    $forecast->{$station}{PROBPRECIP}[$i] and
        print ", $forecast->{$station}{PROBPRECIP}[$i]% " .
              "chance of precipitation.";
    print "\n";
  }


=head1 ABSTRACT

Decodes the National Weather Service's CCF data files into a simple Perl
object.

=head1 DESCRIPTION

There's a lot of U.S. weather information available from the National
Weather Service (and the larger National Oceanic and Atmospheric
Administration). There are other CPAN modules for parsing some of this data,
most notably the METAR reports on current aviation weather. And there's
modules for scraping information from various commercial web sites. But, I
couldn't find anything that could read the forecast information made
available in the form of "CCF" -- Coded City Forecast -- files. This
twice-daily forecast data is designed to be processed by computers,
and indeed provides a lot of information in very little space.

Unfortunately, the format is a bit obtuse, so it's handy to have a Perl
module to understand it for you. I couldn't find one, so here's mine.

You can download various CCF files for different locations from
E<lt>http://weather.noaa.gov/pub/data/raw/fp/E<gt>. (Beware -- a long and
sometimes slow directory listing.) Each ccf file (named something like
F<fpus41.kbox.ccf.bos.txt>) contains the entries for several stations. Find
the one that has the station you want to find the forecast for, and feed it
to this module.

You'll get back a hash of the various stations in the report, with each
station containing the forcast data for the next fourteen seven-hour
periods. (Or, sometimes, only thirteen.) You'll have the predicted
condition -- rain, snow, dust, or whatever, high or low temperature, and
chance of precipitation for each time slice.

=head1 USAGE

=head2 Overview

First, create a new Geo::CCF object. Then, feed it CCF data as a single
string. Then, simply look at the data you want to know for the station you
want.

=head2 Functions

=over

=item ccf()

ccf() is the main thing. Give it a CCF file as a string, and
it'll parse it so you can access the forecast data in a sane way.

=back

=head2 Constants

=over

=item NUM_FORECAST_PERIODS

Unless the CCF file format changes and this module is updated, this will be
14. The forecast is for seven days, in 12-hour periods -- that is, 14
half-days. Conceivably in the future, the NWS could extend the format to
cover ten days, or when the orbital weather control satellite network is in
place, whole years at a time.

=back

=head2 Variables

The important thing to realize is that these values are all stored on a
per-station basis -- the top-level hash is an index of all of the stations
decoded from the given CCF data. 

Furthermore, much of the data is contained in 14-element arrays, with each
element corresponding to a future 12-hour period. In order, of course. And,
sometimes the last value is missing -- nothing to be done about that! You
can use the Geo::CCF::NUM_FORECAST_PERIODS constant as a loop index as in
the example above, but in actual practice, it may be nicer to just check the
arrays to see if the forecast for the desired future period is defined.

=over

=item CONDITION

14-element Array containing the predicted predominant weather. Possibilities
are:

=over

=item * Fair (A)

=item * Partly Cloudy (B)

=item * Cloudy (C)

=item * Dust (D)

=item * Mostly Cloudy (E)

=item * Foggy (F)

=item * Very Hot (G)

=item * Haze (H)

=item * Very Cold (I)

=item * Snow Showers (J)

=item * Smoke (K)

=item * Drizzle (L)

=item * Snow Showers (M)

=item * Windy (N)

=item * Rain/Snow Mix (O)

=item * Blizzard (P)

=item * Blowing Snow (Q)

=item * Rain (R)

=item * Snow (S)

=item * Thunderstorms (T)

=item * Sunny (U)

=item * Clear (V)

=item * Rain Showers (W)

=item * Sleet (X)

=item * Freezing Rain (Y)

=item * Freezing Drizzle (Z)

=back

You won't see the parenthetical letter in the results -- that's the code
in the original file. Some are mnemonic (C or R or S or T) and some are not
quite so (J or P or pretty much all the rest). Anyway, you have this module
so you don't have to worry about that.

Also of note: the source data is in 24 hour periods, but is duplicated by
here so that it matches the temperature and precipitation forecast. Don't be
fooled.

=item HIGH and LOW

14-element arrays containing predicted temperature information in degrees
Fahrenheit. Only one of HIGH or LOW will have a defined value for a given
time period -- generally, LOW at night and HIGH during the day. I realize
that this is stupid and annoying, but that's the source data.

=item PROBPRECIP

14-element array containing the probability (as an integer 0-100) of
precipitation during the 12-hour period.

=item CCFCODE

Each file starts with a six character alphanumeric identifier. Presumably,
this means something to the National Weather Service, but it means nothing
to me. If you pass this file header info to the ccf() method, it'll put that
info here. Otherwise, it'll be undefined.

=item CCFSTATION

From the optional file header -- the station which generated the CCF report,
which probably covers many other stations as well. Not really very generally
useful.

=item CCFDATE and CCFTIME

Also from the file header, and not as useful as you may hope. Unfortunately,
the date is only the day of the month in which the report was generated, so
you need to have some greater context to do anything useful. And the time is
in GMT, which isn't bad, but again, requires knowledge of context.

Anyway, CCFDATE is a two-digit number, and CCFTIME is four digits for HHMM.

=item FORECASTERID

A two-digit number representing the actual forecaster. Maybe this could be
used to produce an accuracy scorecard. Otherwise, probably not of general
usefulness.

=back

=head1 BUGS

None that I know, but I haven't done widespread testing. It's possible that
there's some data-format glitches which I haven't covered. If you find any
problems, please e-mail me.

=head1 CAVEATS

No attempt is made to validate incoming data. If you give ccf() garbage, it
will return with nothing, at best. And if you give slightly weird data, you
won't get any clue back about what's wrong. That ought to be fixed.

The two-line format is apparently an expansion of the earlier one line. This
module won't cope with missing second lines.


There also aren't any test cases. Tests are good to have.

=head1 SEE ALSO

Web site: E<lt>http://www.mattdm.org/misc/perl-Geo-CCF/E<gt>

CCF info:
E<lt>http://www.crh.noaa.gov/lmk/product_guide/products/forecast/ccf.htmE<gt>

Other modules: Geo::METAR, Geo::WeatherNOAA, Geo::Weather

=head1 AUTHOR

Matthew Miller, E<lt>mattdm@mattdm.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2004 by Matthew Miller

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
